from __future__ import annotations

from typing import Any, Dict, Optional, Union

import discord

from .interaction import *
from .types import OptionType

__all__ = (
    "InteractionDataOption",
    "ApplicationCommandInteractionData",
    "SlashInteractionData",
    "ContextMenuInteractionData",
    "SlashInteraction",
    "ContextMenuInteraction",
    "Interaction",
)


class Resolved:
    def __init__(self, *, data, guild, state):
        self.members: Dict[int, discord.Member] = {}
        self.users: Dict[int, discord.User] = {}
        self.roles: Dict[int, discord.Role] = {}
        self.channels: Dict[int, discord.abc.GuildChannel] = {}
        self.messages: Dict[int, discord.Message] = {}

        users = data.get("users", {})
        members = data.get("members", {})
        roles = data.get("roles", {})
        channels = data.get("channels", {})
        messages = data.get("messages", {})

        for ID, data in users.items():
            user_id = int(ID)
            if ID in members:
                self.members[user_id] = discord.Member(
                    data={**members[ID], "user": data}, guild=guild, state=state  # type: ignore
                )
            else:
                self.users[user_id] = discord.User(state=state, data=data)

        for ID, data in roles.items():
            self.roles[int(ID)] = discord.Role(guild=guild, state=state, data=data)

        for ID, data in channels.items():
            data["position"] = 0
            factory, ch_type = discord.channel._channel_factory(data["type"])
            if factory:
                self.channels[int(ID)] = factory(guild=guild, data=data, state=state)

        for ID, data in messages.items():
            channel_id = int(data["channel_id"])
            channel = guild.get_channel(channel_id) if guild else None
            if channel is None:
                channel = state.get_channel(channel_id)
            self.messages[int(ID)] = discord.Message(state=state, channel=channel, data=data)

    def __repr__(self):
        return (
            "<Resolved users={0.users!r} members={0.members!r} "
            "roles={0.roles!r} channels={0.channels!r} "
            "messages={0.messages!r}>"
        ).format(self)

    def get(self, any_id: Any) -> Optional[discord.abc.Snowflake]:
        any_id = int(any_id)
        if any_id in self.members:
            return self.members[any_id]
        if any_id in self.users:
            return self.users[any_id]
        if any_id in self.roles:
            return self.roles[any_id]
        if any_id in self.channels:
            return self.channels[any_id]
        if any_id in self.messages:
            return self.messages[any_id]
        
        return None


class ApplicationCommandInteractionData:
    def __init__(self, *, data, guild, state):
        self.id = int(data["id"])
        self.type: int = data["type"]
        self.name: str = data["name"]
        self.resolved = Resolved(data=data.get("resolved", {}), guild=guild, state=state)


class InteractionDataOption:
    """
    Represents user's input for a specific option

    Attributes
    ----------
    name : str
        The name of the option
    value : Any
        The value of the option
    options : dict
        | Represents options of a sub-slash-command.
        | {``name``: :class:`InteractionDataOption`, ...}
    """

    def __init__(self, *, data, resolved: Resolved):
        self.name: str = data["name"]
        self.type: int = data["type"]
        self.value: Any = data.get("value")
        # Type 6 and higher requires resolved data
        if self.type == OptionType.USER:
            self.value = int(self.value)
            if self.value in resolved.members:
                self.value = resolved.members[self.value]
            elif self.value in resolved.users:
                self.value = resolved.users[self.value]
        elif self.type == OptionType.CHANNEL:
            self.value = int(self.value)
            if self.value in resolved.channels:
                self.value = resolved.channels[self.value]
        elif self.type == OptionType.ROLE:
            self.value = int(self.value)
            if self.value in resolved.roles:
                self.value = resolved.roles[self.value]
        elif self.type == OptionType.MENTIONABLE:
            self.value = int(self.value)
            self.value = resolved.get(self.value, self.value)
        # Converting sub options
        self.options: Dict[str, InteractionDataOption] = {
            o["name"]: InteractionDataOption(data=o, resolved=resolved) for o in data.get("options", [])
        }

    def __repr__(self):
        return "<InteractionDataOption name='{0.name}' value={0.value} options={0.options}>".format(self)

    def _to_dict_values(self, connectors: Dict[str, str] = None) -> Dict[str, Any]:
        connectors = connectors or {}
        out = {}
        for kw, val in self.options.items():
            new_kw = connectors.get(kw, kw)
            out[new_kw] = val.value if val.type > 2 else val
        return out

    @property
    def sub_command(self) -> Optional[InteractionDataOption]:
        opt = self.option_at(0)
        if opt is not None and opt.type == 1:
            return opt
        
        return None

    def get_option(self, name: str) -> Optional[InteractionDataOption]:
        """
        Get the raw :class:`InteractionDataOption` matching the specified name

        Parameters
        ----------
        name : str
            The name of the option you want to get

        Returns
        -------
        option : InteractionDataOption | ``None``
        """
        return self.options.get(name)

    def get(self, name: str, default: Any = None) -> Union[InteractionDataOption, Any]:  # put T here maybe
        """
        Get the value of an option with the specified name

        Parameters
        ----------
        name : str
            the name of the option you want to get
        default : any
            what to return in case nothing was found

        Returns
        -------
        option_value : any
            The option type isn't ``SUB_COMMAND_GROUP`` or ``SUB_COMMAND``
        option: InteractionDataOption | ``default``
            Otherwise
        """
        for n, o in self.options.items():
            if n == name:
                return o.value if o.type > 2 else o
        return default

    def option_at(self, index: int) -> Optional[InteractionDataOption]:
        """Similar to :class:`InteractionData.option_at`"""
        return list(self.options.values())[index] if 0 <= index < len(self.options) else None


class SlashInteractionData(ApplicationCommandInteractionData):
    """
    Attributes
    ----------
    id : :class:`int`
        The id of the interaction
    name : :class:`str`
        The name of activated slash-command
    options : :class:`dict`
        | Represents options of the slash-command.
        | {``name``: :class:`InteractionDataOption`, ...}
    resolved : :class:`Resolved`
        The collection of related objects, such as users, members, roles, channels and messages
    """

    def __init__(self, *, data, guild, state):
        super().__init__(data=data, guild=guild, state=state)
        self.options = {
            o["name"]: InteractionDataOption(data=o, resolved=self.resolved) for o in data.get("options", [])
        }

    def __repr__(self) -> str:
        return "<SlashInteractionData id={0.id} type={0.type} name={0.name!r} options={0.options!r}>".format(self)

    def __getitem__(self, key: str) -> Optional[InteractionDataOption]:
        if isinstance(key, str):
            opt = self.get_option(key)
        elif isinstance(key, int):
            opt = self.option_at(key)
        else:
            raise TypeError(f"unsupported key type. Expected str or int, but received {type(key)} instead")
        if opt is None:
            return None
        return opt.value if opt.type > 2 else opt

    def _to_dict_values(self, connectors: Dict[str, str] = None) -> Dict[str, Any]:
        connectors = connectors or {}
        out = {}
        for kw, val in self.options.items():
            new_kw = connectors.get(kw, kw)
            out[new_kw] = val.value if val.type > 2 else val._to_dict_values(connectors)
        return out

    def _wrap_choices(self, slash_command):
        def recursive_wrapper(wrapped_data, parent):
            for option in parent.options:
                data_option = wrapped_data.get_option(option.name)
                if data_option is None:
                    continue
                if len(option._choice_connectors) > 0:
                    data_option.value = option._choice_connectors.get(data_option.value, data_option.value)
                recursive_wrapper(data_option, option)

        recursive_wrapper(self, slash_command)

    @property
    def sub_command(self) -> Optional[InteractionDataOption]:
        opt = self.option_at(0)
        if opt is not None and opt.type == 1:
            return opt

    @property
    def sub_command_group(self) -> Optional[InteractionDataOption]:
        opt = self.option_at(0)
        if opt is not None and opt.type == 2:
            return opt

    def get_option(self, name: str) -> Optional[InteractionDataOption]:
        """
        Get the raw :class:`InteractionDataOption` matching the specified name

        Parameters
        ----------
        name : str
            The name of the option you want to get

        Returns
        -------
        option : :class:`InteractionDataOption` | ``None``
        """
        return self.options.get(name)

    def get(self, name: str, default: Any = None) -> Union[InteractionDataOption, Any]:
        """
        Get the value of an option with the specified name

        Parameters
        ----------
        name : str
            the name of the option you want to get
        default : any
            what to return in case nothing was found

        Returns
        -------
        option_value : any
            The option type isn't ``SUB_COMMAND_GROUP`` or ``SUB_COMMAND``
        option: :class:`InteractionDataOption` | ``default``
            Otherwise
        """
        opt = self.options.get(name)
        if opt is None:
            return default
        return opt.value if opt.type > 2 else opt

    def option_at(self, index: int) -> Optional[InteractionDataOption]:
        """
        Get an option by it's index

        Parameters
        ----------
        index : int
            the index of the option you want to get

        Returns
        -------
        option : :class:`InteractionDataOption` | ``None``
            the option located at the specified index
        """
        return list(self.options.values())[index] if 0 <= index < len(self.options) else None


class ContextMenuInteractionData(ApplicationCommandInteractionData):
    def __init__(self, data, guild, state):
        super().__init__(data=data, guild=guild, state=state)
        self.target_id = int(data["target_id"])
        self.target = self.resolved.get(self.target_id)

    def __repr__(self):
        return (
            f"<ContextMenuInteractionData id={self.id} type={self.type} " f"name={self.name!r} target={self.target!r}>"
        )

    @property
    def member(self):
        if isinstance(self.target, discord.Member):
            return self.target

    @property
    def user(self):
        if isinstance(self.target, discord.abc.User):
            return self.target

    @property
    def message(self):
        if isinstance(self.target, discord.Message):
            return self.target


class SlashInteraction(BaseInteraction):
    """
    Every interaction with slash-commands is represented by instances of this class

    Attributes
    ----------
    author : :class:`discord.Member` | :class:`discord.User`
        The member/user that used the slash-command.
    guild : discord.Guild
        The guild where interaction was created
    channel : :class:`discord.TextChannel`
        The channel where interaction was created
    data : :class:`InteractionData`
        The arguments that were passed
    created_at : :class:`datetime.datetime`
        Then interaction was created
    expired : :class:`bool`:
        Whether the interaction token is still valid
    """

    def __init__(self, client, payload):
        super().__init__(client, payload)

        state = client._connection
        self.prefix: str = "/"  # Just in case
        self.data = SlashInteractionData(data=payload.get("data", {}), guild=self.guild, state=state)
        self.invoked_with = self.data.name
        # what's this???
        self.slash_command: Optional[Any] = None
        self.sub_command_group: Optional[Any] = None
        self.sub_command: Optional[Any] = None

    def __repr__(self) -> str:
        return (
            "<SlashInteraction id={0.id} version={0.version} type={0.type} "
            "token='{0.token}' guild={0.guild} channel={0.channel} "
            "author={0.author} data={0.data!r}>"
        ).format(self)

    def __getitem__(self, key):
        return self.data[key]

    def _wrap_choices(self, slash_command):
        self.data._wrap_choices(slash_command)

    def get(self, name: str, default: Any = None):
        """Equivalent to :class:`InteractionData.get`"""
        return self.data.get(name, default)

    def get_option(self, name: str):
        """Equivalent to :class:`InteractionData.get_option`"""
        return self.data.get_option(name)

    def option_at(self, index: int):
        """Equivalent to :class:`InteractionData.option_at`"""
        return self.data.option_at(index)


class ContextMenuInteraction(BaseInteraction):
    def __init__(self, client, payload):
        super().__init__(client, payload)
        self.data = ContextMenuInteractionData(data=payload.get("data", {}), guild=self.guild, state=client._connection)
        self.user_command: Optional[Any] = None
        self.message_command: Optional[Any] = None

    def __repr__(self):
        return (
            f"<ContextMenuInteraction id={self.id} guild={self.guild!r} "
            f"channel={self.channel!r} author={self.author!r} data={self.data!r}>"
        )

    @property
    def target(self):
        return self.data.target

    @property
    def user(self):
        return self.data.user

    @property
    def member(self):
        return self.data.member

    @property
    def message(self):
        return self.data.message


Interaction = SlashInteraction



import asyncio
import datetime
import functools
import inspect
from enum import EnumMeta
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

import discord
from discord.ext.commands import BucketType, Cooldown, CooldownMapping
from discord.ext.commands.errors import ConversionError

from ..interactions import BaseInteraction, SlashInteraction
from ..interactions.application_command import OptionParam
from .errors import *

__all__ = (
    "BucketType",
    "InvokableApplicationCommand",
    "check",
    "check_any",
    "has_role",
    "has_any_role",
    "bot_has_role",
    "bot_has_any_role",
    "has_permissions",
    "bot_has_permissions",
    "has_guild_permissions",
    "bot_has_guild_permissions",
    "dm_only",
    "guild_only",
    "is_owner",
    "is_nsfw",
    "cooldown",
)


class InvokableApplicationCommand:
    def __init__(self, func: Callable[..., Awaitable], *, name: str = None, **kwargs):
        self.func = func
        self.name = name or func.__name__
        self._error_handler: Optional[Callable[..., Any]] = None
        self.auto_sync: bool = True
        # Extract checks
        if hasattr(func, "__slash_checks__"):
            self.checks: List[Callable[..., Any]] = func.__slash_checks__
        else:
            self.checks: List[Callable[..., Any]] = []
        # Cooldown
        try:
            cooldown = func.__slash_cooldown__
        except AttributeError:
            cooldown = None
        if cooldown is None:
            try:
                # Assuming that it's discord.py 1.7.0+
                self._buckets = CooldownMapping(cooldown, BucketType.default)
            except Exception:
                # discord.py <= 1.6.x
                try:
                    self._buckets = CooldownMapping(cooldown)  # type: ignore
                except Exception:
                    # Hopefully we never reach this
                    self._buckets: CooldownMapping = None  # type: ignore
        else:
            self._buckets = cooldown
        # Add custom kwargs
        for kw, value in kwargs.items():
            if not hasattr(self, kw):
                setattr(self, kw, value)

    async def __call__(self, *args, **kwargs):
        return await self.func(*args, **kwargs)

    @staticmethod
    def _isinstance(obj: Any, typ: type) -> bool:
        if not isinstance(typ, type) or isinstance(typ, EnumMeta):  # special annotation
            return True
        if issubclass(typ, discord.User):
            return isinstance(obj, (discord.User, discord.Member, discord.ClientUser))
        else:
            return isinstance(obj, typ)

    def _prepare_cooldowns(self, inter: Any):
        if self._buckets.valid:
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(inter, current)
            retry_after = bucket.update_rate_limit(current)
            if retry_after:
                raise CommandOnCooldown(bucket, retry_after)

    def _dispatch_error(self, cog: Any, inter: BaseInteraction, error: Exception):
        asyncio.create_task(self._invoke_error_handler(cog, inter, error))

    async def _run_checks(self, inter: BaseInteraction):
        for _check in self.checks:
            if not await _check(inter):
                raise InteractionCheckFailure(f"command <{self.name}> has failed")

    async def _maybe_cog_call(self, cog: Any, inter: BaseInteraction):
        if cog:
            return await self(cog, inter)
        else:
            return await self(inter)

    async def _invoke_error_handler(self, cog: Any, inter: BaseInteraction, error: Exception):
        if self._error_handler is None:
            return
        if cog:
            await self._error_handler(cog, inter, error)
        else:
            await self._error_handler(inter, error)

    def error(self, func: Callable[..., Awaitable]):
        """
        A decorator that makes the function below
        work as error handler for this command.
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("The local error handler must be an async function")
        self._error_handler = func
        return func

    def is_on_cooldown(self, inter: Any):
        """
        Checks whether the slash command is currently on cooldown.

        Parameters
        -----------
        inter: :class:`SlashInteraction`
            The interaction to use when checking the commands cooldown status.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the slash command is on cooldown.
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(inter)
        dt = inter.created_at
        current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        return bucket.get_tokens(current) == 0

    def reset_cooldown(self, inter: Any):
        """
        Resets the cooldown on this slash command.

        Parameters
        -----------
        inter: :class:`SlashInteraction`
            The interaction to reset the cooldown under.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)
            bucket.reset()

    def get_cooldown_retry_after(self, inter: Any):
        """
        Retrieves the amount of seconds before this slash command can be tried again.

        Parameters
        -----------
        inter: :class:`SlashInteraction`
            The interaction to retrieve the cooldown from.

        Returns
        --------
        :class:`float`
            The amount of time left on this slash command's cooldown in seconds.
            If this is ``0.0`` then the slash command isn't on cooldown.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(inter)
            dt = inter.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            return bucket.get_retry_after(current)

        return 0.0


def class_name(func: Callable) -> Optional[str]:
    res = func.__qualname__[: -len(func.__name__)]
    return None if len(res) == 0 else res[:-1]


def get_class(func: Callable) -> Any:
    if inspect.isfunction(func):
        cn = class_name(func)
        if cn is None:
            return None
        return getattr(inspect.getmodule(func), cn, None)


def check(predicate: Callable[..., Any]):
    """
    A function that converts ``predicate(interaction)`` functions
    into application command decorators

    Example

    ::

        def is_guild_owner():
            def predicate(inter):
                return inter.author.id == inter.guild.owner_id
            return check(predicate)

        @is_guild_owner()
        @slash.command(description="Says Hello if you own the guild")
        async def hello(inter):
            await inter.reply("Hello, Mr.Owner!")

    """
    if inspect.iscoroutinefunction(predicate):
        wrapper = predicate  # type: ignore
    else:

        async def wrapper(ctx):
            return predicate(ctx)

    def decorator(func):
        if isinstance(func, InvokableApplicationCommand):
            func.checks.append(wrapper)
        else:
            if not hasattr(func, "__slash_checks__"):
                func.__slash_checks__ = []
            func.__slash_checks__.append(wrapper)
        return func

    decorator.predicate = wrapper
    return decorator


def check_any(*checks: Callable[..., Any]):
    """Similar to ``commands.check_any``"""

    unwrapped = []
    for wrapped in checks:
        try:
            pred = wrapped.predicate
        except AttributeError:
            raise TypeError("%r must be wrapped by commands.check decorator" % wrapped) from None
        else:
            unwrapped.append(pred)

    async def predicate(ctx):
        errors = []
        for func in unwrapped:
            try:
                value = await func(ctx)
            except SlashCheckFailure as e:
                errors.append(e)
            else:
                if value:
                    return True
        # if we're here, all checks failed
        raise CheckAnyFailure(unwrapped, errors)

    return check(predicate)


def has_role(item: Union[int, str]):
    """Similar to ``commands.has_role``"""

    def predicate(ctx):
        if not isinstance(ctx.channel, discord.abc.GuildChannel):
            raise NoPrivateMessage()

        if isinstance(item, int):
            role = discord.utils.get(ctx.author.roles, id=item)
        else:
            role = discord.utils.get(ctx.author.roles, name=item)
        if role is None:
            raise MissingRole(item)
        return True

    return check(predicate)


def has_any_role(*items: Union[int, str]):
    """Similar to ``commands.has_any_role``"""

    def predicate(ctx):
        if not isinstance(ctx.channel, discord.abc.GuildChannel):
            raise NoPrivateMessage()

        getter = functools.partial(discord.utils.get, ctx.author.roles)
        if any(
            getter(id=item) is not None if isinstance(item, int) else getter(name=item) is not None for item in items
        ):
            return True
        raise MissingAnyRole(items)

    return check(predicate)


def bot_has_role(item: Union[int, str]):
    """Similar to ``commands.bot_has_role``"""

    def predicate(ctx):
        ch = ctx.channel
        if not isinstance(ch, discord.abc.GuildChannel):
            raise NoPrivateMessage()

        me = ch.guild.me
        if isinstance(item, int):
            role = discord.utils.get(me.roles, id=item)
        else:
            role = discord.utils.get(me.roles, name=item)
        if role is None:
            raise BotMissingRole(item)
        return True

    return check(predicate)


def bot_has_any_role(*items: Union[int, str]):
    """Similar to ``commands.bot_has_any_role``"""

    def predicate(ctx):
        ch = ctx.channel
        if not isinstance(ch, discord.abc.GuildChannel):
            raise NoPrivateMessage()

        me = ch.guild.me
        getter = functools.partial(discord.utils.get, me.roles)
        if any(
            getter(id=item) is not None if isinstance(item, int) else getter(name=item) is not None for item in items
        ):
            return True
        raise BotMissingAnyRole(items)

    return check(predicate)


def has_permissions(**perms: bool):
    """Similar to ``commands.has_permissions``"""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError("Invalid permission(s): %s" % (", ".join(invalid)))

    def predicate(ctx):
        ch = ctx.channel
        permissions = ch.permissions_for(ctx.author)

        missing = [perm for perm, value in perms.items() if getattr(permissions, perm) != value]

        if not missing:
            return True

        raise MissingPermissions(missing)

    return check(predicate)


def bot_has_permissions(**perms: bool):
    """Similar to ``commands.bot_has_permissions``"""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError("Invalid permission(s): %s" % (", ".join(invalid)))

    def predicate(ctx):
        guild = ctx.guild
        me = guild.me if guild is not None else ctx.bot.user
        permissions = ctx.channel.permissions_for(me)

        missing = [perm for perm, value in perms.items() if getattr(permissions, perm) != value]

        if not missing:
            return True

        raise BotMissingPermissions(missing)

    return check(predicate)


def has_guild_permissions(**perms: bool):
    """Similar to ``commands.has_guild_permissions``"""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError("Invalid permission(s): %s" % (", ".join(invalid)))

    def predicate(ctx):
        if not ctx.guild:
            raise NoPrivateMessage

        permissions = ctx.author.guild_permissions
        missing = [perm for perm, value in perms.items() if getattr(permissions, perm) != value]

        if not missing:
            return True

        raise MissingPermissions(missing)

    return check(predicate)


def bot_has_guild_permissions(**perms: bool):
    """Similar to ``commands.bot_has_guild_permissions``"""

    invalid = set(perms) - set(discord.Permissions.VALID_FLAGS)
    if invalid:
        raise TypeError("Invalid permission(s): %s" % (", ".join(invalid)))

    def predicate(ctx):
        if not ctx.guild:
            raise NoPrivateMessage

        permissions = ctx.guild.me.guild_permissions
        missing = [perm for perm, value in perms.items() if getattr(permissions, perm) != value]

        if not missing:
            return True

        raise BotMissingPermissions(missing)

    return check(predicate)


def dm_only():
    """Similar to ``commands.dm_only``"""

    def predicate(ctx):
        if ctx.guild is not None:
            raise PrivateMessageOnly()
        return True

    return check(predicate)


def guild_only():
    """Similar to ``commands.guild_only``"""

    def predicate(ctx):
        if ctx.guild is None:
            raise NoPrivateMessage()
        return True

    return check(predicate)


def is_owner():
    """Similar to ``commands.is_owner``"""

    async def predicate(ctx):
        if not await ctx.bot.is_owner(ctx.author):
            raise NotOwner("You do not own this bot.")
        return True

    return check(predicate)


def is_nsfw():
    """Similar to ``commands.is_nsfw``"""

    def pred(ctx):
        ch = ctx.channel
        if ctx.guild is None or (isinstance(ch, discord.TextChannel) and ch.is_nsfw()):
            return True
        raise NSFWChannelRequired(ch)

    return check(pred)


def cooldown(rate: int, per: float, type: BucketType = BucketType.default):
    """
    A decorator that adds a cooldown to a slash-command. Similar to **discord.py** cooldown decorator.

    A cooldown allows a command to only be used a specific amount
    of times in a specific time frame. These cooldowns can be based
    either on a per-guild, per-channel, per-user, per-role or global basis.
    Denoted by the third argument of ``type`` which must be of enum
    type ``BucketType``.

    If a cooldown is triggered, then ``CommandOnCooldown`` is triggered in
    ``on_slash_command_error`` in the local error handler.

    A command can only have a single cooldown.

    Parameters
    ----------
    rate : int
        The number of times a command can be used before triggering a cooldown.
    per : float
        The amount of seconds to wait for a cooldown when it's been triggered.
    type : BucketType
        The type of cooldown to have.
    """

    def decorator(func):
        try:
            cooldown_obj = Cooldown(rate, per, type)  # type: ignore
        except Exception:
            cooldown_obj = Cooldown(rate, per)

        try:
            mapping = CooldownMapping(cooldown_obj)  # type: ignore
        except Exception:
            mapping = CooldownMapping(cooldown_obj, type)

        if isinstance(func, InvokableApplicationCommand):
            func._buckets = mapping
        else:
            func.__slash_cooldown__ = mapping

        return func

    return decorator



import asyncio
from datetime import datetime, timedelta
from typing import Any, Awaitable, Callable, Dict, List, Tuple

import discord

from ..interactions import MessageInteraction

__all__ = ("ClickListener", "ClickManager")


PER_MESSAGE_LISTENERS: Dict[int, 'ClickListener'] = {}


async def _on_button_click(inter):
    redir = PER_MESSAGE_LISTENERS.get(inter.message.id)
    if redir is not None:
        redir._toggle_listeners(inter)


class ClickListener:
    """
    Creates a nice click manager for a message. You can use this class,
    or call :class:`discord.Message.create_click_listener(timeout)`
    in order to create an instance.

    Click manager allows to process button clicks under a specific message
    using decorator-based interface.

    Parameters
    ----------
    message_id : :class:`int`
        the ID of the message with buttons
    timeout : :class:`float`
        the amount of seconds required after the last matched interaction
        for the click manager to finish working.
        Defaults to ``None``, which means no timeout.
    """

    __slots__ = ("id", "_listeners", "_timeout_waiter", "_timeout", "_ends_at")

    def __init__(self, message_id: int, timeout: float = 0):
        self.id = message_id
        # listener, condition, cancel_others, reset_timeout
        self._listeners: List[Tuple[Callable[[MessageInteraction], Awaitable[Any]], Callable[[MessageInteraction], bool], bool, bool]] = []
        self._timeout_waiter = None
        self._timeout = timeout
        PER_MESSAGE_LISTENERS[message_id] = self
        # Launch a finishing task
        if timeout is not None:
            self._ends_at = datetime.now() + timedelta(seconds=timeout)
            asyncio.create_task(self._wait_and_finish())
        else:
            self._ends_at = None

    async def _wait_and_finish(self):
        delay = self._timeout
        while True:
            await asyncio.sleep(delay)
            now = datetime.now()
            if self._ends_at and self._ends_at > now:
                delay = (self._ends_at - now).total_seconds()
            else:
                break
        PER_MESSAGE_LISTENERS.pop(self.id, None)
        if self._timeout_waiter is not None:
            await self._timeout_waiter()

    def _toggle_listeners(self, inter: MessageInteraction):
        task_toggled = False
        for listener, condition, cancel_others, reset_timeout in self._listeners:
            try:
                res = condition(inter)
            except Exception:
                res = False
            if res:
                task_toggled = task_toggled or reset_timeout
                asyncio.create_task(listener(inter))
                if cancel_others:
                    break
        # Delay more
        if task_toggled and self._ends_at is not None:
            self._ends_at = datetime.now() + timedelta(seconds=self._timeout)

    def kill(self):
        """
        Kills the click manager. Only useful if the ``timeout``
        param was specified as ``None``.
        """
        self._timeout_waiter = None  # Also kills the timeout waiter
        PER_MESSAGE_LISTENERS.pop(self.id, None)

    def timeout(self, func: Callable[..., Any]):
        """
        A decorator that makes the function below waiting for click listener timeout.
        """
        if not asyncio.iscoroutinefunction(func):
            async def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            new_func = wrapper
        else:
            new_func = func
        self._timeout_waiter = new_func
        return func

    def matching_condition(self, check: Callable[[MessageInteraction], bool], *, cancel_others: bool = False, reset_timeout: bool = True):
        """
        A decorator that makes the function below waiting for a click
        matching the specified conditions.

        Parameters
        ----------
        check : :class:`function`
            the required condition. This function must take exactly one argument
            which is guaranteed to be an instance of :class:`MessageInteraction`.
            This function must return ``True`` or ``False``.
        cancel_others : :class:`bool`
            defaults to ``False``. Specifies
            whether to cancel all other local listeners or not.
            For example, if this parameter is ``False``, the library will
            activate all other listeners matching the interaction,
            untill all listeners are toggled or some of them cancels others.
        reset_timeout : :class:`bool`
            defaults to ``True``. Specifies whether to restart the timer or not.
            By restarting the timer, you extend the lifespan of all local listeners.
        """
        def deco(func):
            if not asyncio.iscoroutinefunction(func):
                async def wrapper(*args, **kwargs):
                    return func(*args, **kwargs)
                new_func = wrapper
            else:
                new_func = func
            self._listeners.append((new_func, check, cancel_others, reset_timeout))
            return func
        return deco

    def from_user(self, user: discord.User, *, cancel_others: bool = False, reset_timeout: bool = True):
        """
        A decorator that makes the function below waiting for a click
        from the specified user.

        Parameters are the same as in :meth:`matching_condition`, except
        ``check`` parameter is replaced with a ``user`` to compare with.
        """
        def is_user(inter: MessageInteraction):
            return inter.author == user
        return self.matching_condition(
            is_user,
            cancel_others=cancel_others,
            reset_timeout=reset_timeout
        )

    def not_from_user(self, user: discord.User, *, cancel_others: bool = False, reset_timeout: bool = True):
        """
        A decorator that makes the function below waiting for a click
        from a user not maching the specified one.

        Parameters are the same as in :meth:`matching_condition`, except
        ``check`` parameter is replaced with a ``user`` to compare with.
        """
        def is_not_user(inter):
            return inter.author != user
        return self.matching_condition(
            is_not_user,
            cancel_others=cancel_others,
            reset_timeout=reset_timeout
        )

    def no_checks(self, *, cancel_others: bool = False, reset_timeout: bool = True):
        """
        A decorator that makes the function below waiting for any click.

        Parameters are the same as in :meth:`matching_condition`, except
        there's no ``check``.
        """
        def always(inter):
            return True
        return self.matching_condition(
            always,
            cancel_others=cancel_others,
            reset_timeout=reset_timeout
        )

    def matching_id(self, custom_id: str, *, cancel_others: bool = False, reset_timeout: bool = True):
        """
        A decorator that makes the function below waiting for a click
        of the button matching the specified custom_id.

        Parameters are the same as in :meth:`matching_condition`, except
        ``check`` parameter is replaced with ``custom_id``.
        """
        check = lambda inter: inter.component.custom_id == custom_id
        return self.matching_condition(
            check,
            cancel_others=cancel_others,
            reset_timeout=reset_timeout
        )


ClickManager = ClickListener
